"""Define the platereader class."""

import warnings
from pathlib import Path

import omniplate
import omniplate.clogger as clogger


class platereader:
    """
    To analyse plate-reader data.

    Data are corrected for autofluorescence and growth rates found using
    a Gaussian process.

    All data is stored used Panda's dataframes and plotted using Seaborn.

    Three dataframes are created. If p is an instance of the platereader class,
    then p.r contains the raw data for each well in the plate; p.s contains the
    processed time-series using the data from all relevant wells; and p.sc
    contains any summary statistics, such as 'max gr'.

    For time series sampled from a Gaussian process, the mean is used as the
    statistics and errors are estimated by the standard deviation.
    For statistics calculated from time series, the median is used and errors
    are estimated by half the interquartile range, with the distribution of
    the statistic calculated by sampling time series.

    Examples
    -------
    A typical work flow is:

    >>> import omniplate as om

    then either

    >>> p= om.platereader('GALdata.xlsx', 'GALcontents.xlsx',
    ...                    wdir= 'data/')

    or

    >>> p= om.platereader()
    >>> p.load('GALdata.xls', 'GALcontents.xlsx')

    and to analyse OD data

    >>> p.plot('OD', plate= True)
    >>> p.correctOD()
    >>> p.correctmedia()
    >>> p.plot(y= 'OD')
    >>> p.plot(y= 'OD', hue= 'strain',
    ...        conditionincludes= ['Gal', 'Glu'],
    ...        strainexcludes= 'HXT7')
    >>> p.getstats('OD')

    and for fluorescence data

    >>> p.correctauto(['GFP', 'AutoFL'])
    >>> p.plot(y= 'cGFPperOD', hue= 'condition')

    and to save the results

    >>> p.savefigs()
    >>> p.exportdf()

    General properties of the data and of previous processing are shown with:

    >>> p.info
    >>> p.attributes()
    >>> p.corrections()
    >>> p.log

    See also
        http://swainlab.bio.ed.ac.uk/software/omniplate/index.html
    for a tutorial, which can be opened directly using

    >>> p.webhelp()
    """

    # ratio of fluorescence emission at 585nm to emiisions at 525nm for eGFP
    _gamma = 0.114  # TODO delete

    def __init__(
        self,
        dnames=None,
        anames=None,
        experimenttype=None,
        wdir=".",
        datadir=".",
        platereadertype="Tecan",
        dsheets=False,
        asheets=False,
        ODfname=None,
        info=True,
        ls=True,
    ):
        """
        Initiate and potentially immediately load data for processing.

        Parameters
        ----------
        dnames: string or list of strings, optional
            The name of the file containing the data from the plate reader or
            a list of file names.
        anames: string or list of strings, optional
            The name of file containing the corresponding annotation or a list
            of file names.
        experimenttype: string or list of strings, optional
            If specified, creates a new experiment_type column in each
            dataframe.
        wdir: string, optional
            The working directory where new files should be saved.
        datadir: string, optional
             The directory containing the data files.
        platereadertype: string
            The type of plate reader, currently either "Tecan" or "tidy" for
            data parsed into a tsv file of columns, with headings "time",
            "well", "OD", "GFP", and any other measurements taken.
        dsheets: list of integers or strings, optional
            The relevant sheets of the Excel files storing the data.
        asheets: list of integers or strings, optional
            The relevant sheets of the corresponding Excel files for the
            annotation.
        info: boolean
            If True (default), display summary information on the data once
            loaded.
        ls: boolean
            If True (default), display contents of working directory.
        """
        # absolute path
        self.wdirpath = Path(wdir)
        self.datadirpath = Path(datadir)
        # enable logging
        self.logger, self.logstream = clogger.initlog(omniplate.__version__)
        if True:
            # warning generated occasionally when sampling from the Gaussian
            # process likely because of numerical errors
            warnings.simplefilter("ignore", RuntimeWarning)
        # dictionary recording extent of analysis
        self.progress = {
            "ignoredwells": {},
            "negativevalues": {},
        }
        self.allexperiments = []
        self.allconditions = {}
        self.allstrains = {}
        self.datatypes = {}
        self.allstrainsconditions = {}
        self.find_available_data
        if dnames is None:
            # list all files in current directory
            if ls:
                self.ls
        else:
            # immediately load data
            self.load(
                dnames=dnames,
                anames=anames,
                experimenttype=experimenttype,
                platereadertype=platereadertype,
                dsheets=dsheets,
                asheets=asheets,
                info=info,
            )
        self.combined = "__combined__"

    ###
    # import methods
    ###
    from .admin import cols_to_underscore, drop
    from .corrections import (
        correctauto,
        correctauto_l,
        correctODformedia,
        correctOD,
    )
    from .getstats import getstats
    from .loaddata import load
    from .manipulatedfs import (
        add_to_sc,
        addcolumn,
        addcommonvar,
        addnumericcolumn,
        getdataframe,
        rename,
        rename_combined,
        restricttime,
        search,
    )
    from .midlog import getmidlog
    from .misc import averageoverexpts, getfitnesspenalty
    from .ominfo import (
        changewdir,
        changedatadir,
        find_available_data,
        info,
        log,
        ls,
        webhelp,
        experiment_map,
    )
    from .omio import exportdf, importdf, savelog, save
    from .omplot import close, plot, savefigs, inspect
    from .omwells import contentsofwells, ignorewells, showwells
    from .promoteractivity import getpromoteractivity

    def __repr__(self):
        """Give standard representation."""
        repstr = f"omniplate.{self.__class__.__name__}: "
        if self.allexperiments:
            for e in self.allexperiments:
                repstr += e + " ; "
            if repstr[-2:] == "; ":
                repstr = repstr[:-3]
        else:
            repstr += "no experiments"
        return repstr


if __name__ == "__main__":
    print(platereader.__doc__)



"""Functions for extracting subsets of the data."""

import numpy as np

import omniplate.omerrors as errors
import omniplate.omgenutils as gu


def getall_labels(self, labeltype, nonull, nomedia):
    """List all the labels available for a given label type."""
    if labeltype == "experiment":
        all_labels = self.allexperiments
    elif labeltype == "condition":
        all_labels = list(
            set(
                [
                    con
                    for e in self.allconditions
                    for con in self.allconditions[e]
                ]
            )
        )
        if nomedia and "media" in all_labels:
            all_labels.pop(all_labels.index("media"))
    elif labeltype == "strain":
        all_labels = list(
            set(
                [
                    strain
                    for e in self.allstrains
                    for strain in self.allstrains[e]
                ]
            )
        )
        if nonull and "Null" in all_labels:
            all_labels.pop(all_labels.index("Null"))
    else:
        raise errors.getsubset("Nothing found.")
    return all_labels


def getset(
    self,
    label,
    labelincludes,
    labelexcludes,
    labeltype,
    nonull,
    nomedia=True,
):
    """Find user-specified list of experiments, conditions, or strains."""
    all_labels = getall_labels(self, labeltype, nonull, nomedia)
    if label != "all":
        # prioritise explicitly specified labels
        labels = gu.makelist(label)
        # check user's choice exists
        bad_labels = [label for label in labels if label not in all_labels]
        # keep only good labels
        labels = [label for label in labels if label not in bad_labels]
    else:
        labels = all_labels
        if labelincludes:
            # find those items containing keywords given in 'includes'
            includes = gu.makelist(labelincludes)
            newset = []
            for s in all_labels:
                gotone = 0
                for item in includes:
                    if item in s:
                        gotone += 1
                if gotone == len(includes):
                    newset.append(s)
            labels = newset
        if labelexcludes:
            # remove any items containing keywords given in 'excludes'
            excludes = gu.makelist(labelexcludes)
            exs = []
            for s in all_labels:
                for item in excludes:
                    if item in s:
                        exs.append(s)
                        break
            for ex in exs:
                if ex in labels:
                    labels.pop(labels.index(ex))
    if labels:
        # sort by numeric values in list entries
        return sorted(labels, key=gu.natural_keys)
    else:
        raise errors.getsubset("No data found.")


def getall(
    self,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
    nonull,
    nomedia=True,
):
    """Return experiments, conditions, and strains."""
    exps = getset(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        "experiment",
        nonull,
        nomedia,
    )
    cons = getset(
        self,
        conditions,
        conditionincludes,
        conditionexcludes,
        "condition",
        nonull,
        nomedia,
    )
    strs = getset(
        self,
        strains,
        strainincludes,
        strainexcludes,
        "strain",
        nonull,
        nomedia,
    )
    return exps, cons, strs


def extractwells(r_df, s_df, experiment, condition, strain, datatypes):
    """
    Extract a list of data matrices from the r dataframe.

    Each column in each matrix contains the data
    from one well.

    Data is returned for each dtype in datatypes, which may include "time", for
    the given experiment, condition, and strain.
    """
    datatypes = gu.makelist(datatypes)
    # restrict time if necessary
    lrdf = r_df[
        (r_df.time >= s_df.time.min()) & (r_df.time <= s_df.time.max())
    ]
    df = lrdf.query(
        "experiment == @experiment and condition == @condition "
        "and strain == @strain"
    )
    if df.empty:
        return None, None
    else:
        # extract time
        dfw_time = (
            df[["time", "well"]]
            .groupby("well", group_keys=True)["time"]
            .apply(list)
        )
        well_times = [dfw_time[well] for well in dfw_time.index]
        # data may have different durations
        longest_i = np.argmax([len(dfw_time[well]) for well in dfw_time.index])
        t = np.array(well_times[longest_i])
        # extract data
        matrices = []
        for dtype in datatypes:
            dfw_dtype = (
                df[[dtype, "well"]]
                .groupby("well", group_keys=True)[dtype]
                .apply(list)
            )
            data = np.nan * np.ones((len(t), dfw_dtype.shape[0]))
            for i, well in enumerate(dfw_dtype.index):
                data[: len(dfw_dtype[well]), i] = np.array(dfw_dtype[well])
            matrices.append(data)
        if len(datatypes) == 1:
            # return array
            return t, matrices[0]
        else:
            # return list of arrays
            return t, matrices



"""Smooth and estimate time derivatives via Gaussian processes."""

import copy
import warnings

import gaussianprocessderivatives as gp
import matplotlib.pylab as plt
import numpy as np
from gaussianprocessderivatives import gaussianprocessException
from scipy.interpolate import interp1d

import omniplate.omgenutils as gu
import omniplate.omstats as omstats


def runfitderiv(
    t,
    d,
    fitvar,
    derivname,
    experiment,
    condition,
    strain,
    bd=False,
    cvfn="matern",
    empirical_errors=False,
    noruns=10,
    exitearly=True,
    noinits=100,
    nosamples=100,
    logs=False,
    negs2nan=True,
    findareas=False,
    plotlocalmax=True,
    printpeakproperties=False,
    linalgmax=5,
    maxdatapts=None,
    **kwargs,
):
    """
    Run fitderiv to smooth and estimate time derivatives for a single data set.

    Parameters
    ----------
    t: array
        An array of times.
    d: array
        An array of measurements of the variable to be fit.
    fitvar: string
        The name of the variable to be fit.
    derivname: string
        The name of the first time derivative of the variable.
    experiment: string
        The name of the experiment of interest.
    condition: string
        The condition of interest.
    strain: string
        The strain of interest.
    ylabels: list of strings
        The labels for the y-axis
    bd: dictionary, optional
        The bounds on the hyperparameters for the Gaussian process.
        For example, bd= {1: [-2,0])} fixes the bounds on the
        hyperparameter controlling flexibility to be 1e-2 and 1e0.
        The default for a Matern covariance function
        is {0: (-5,5), 1: (-4,4), 2: (-5,2)},
        where the first element controls amplitude, the second controls
        flexibility, and the third determines the magnitude of the
        measurement error.
    cvfn: string, optional
        The covariance function used in the Gaussian process, either
        'matern' or 'sqexp' or 'nn'.
    empirical_errors: boolean, optional
        If True, measurement errors are empirically estimated from the
        variance across replicates at each time point and so vary with
        time.
        If False, the magnitude of the measurement error is fit from the
        data assuming that this magnitude is the same at all time points.
    noruns: integer, optional
        The number of attempts made for each fit. Each attempt is made
        with random initial estimates of the hyperparameters within their
        bounds.
    exitearly: boolean, optional
        If True, stop at the first successful fit.
        If False, use the best fit from all successful fits.
    noinits: integer, optional
        The number of random attempts to find a good initial condition
        before running the optimization.
    nosamples: integer, optional
        The number of samples used to calculate errors in statistics by
        bootstrapping.
    logs: boolean, optional
        If True, find the derivative of the log of the data and should be
        True to determine the specific growth rate when dtype= 'OD'.
    negs2nan: boolean, optional
            If logs=True, negs2nan=True causes all negative numbers to be set
            to np.nan; negs2nan=False, causes all negative numbers to be set
            almost zero.
    findareas: boolean, optional
        If True, find the area under the plot of gr vs OD and the area
        under the plot of OD vs time. Setting to True can make getstats
        slow.
    plotlocalmax: boolean, optional
        If True, mark the highest local maxima found, which is used to
        calculate statistics, on any plots.
    printpeakproperties: boolean, optional
        If True, show properties of any local peaks that have found by
        scipy's find_peaks. Additional properties can be specified as
        kwargs and are passed to find_peaks.
    linalgmax: int, optional
        The number of linear algebra errors to tolerate.
    maxdatapts: integer, optional
        If set, sufficiently large data sets with multiple replicates will
        be subsampled at each time point, randomly picking a smaller
        number of replicates, to reduce the number of data points and so
        run times.
    kwargs: passed to scipy's find_peaks.
    """
    print(f"Fitting {fitvar} for {experiment}: {strain} in {condition}")
    # define statnames
    statnames = [
        f"min_{fitvar}",
        f"max_{fitvar}",
        f"range_{fitvar}",
        f"max_{derivname}",
        f"time_of_max_{derivname}",
    ]
    if derivname == "gr":
        # special names when estimating specific growth rate
        statnames += ["doubling_time", "lag_time"]
    else:
        statnames += [
            f"doubling_time_from_{derivname}",
            f"lag_time_from_{derivname}",
        ]
    # call fitderiv
    f = fitderiv(
        t,
        d,
        cvfn=cvfn,
        logs=logs,
        negs2nan=negs2nan,
        noruns=noruns,
        noinits=noinits,
        exitearly=exitearly,
        bd=bd,
        empirical_errors=empirical_errors,
        linalgmax=linalgmax,
        maxdatapts=maxdatapts,
    )
    if f.success:
        # check derivative has been sensibly defined
        if np.max(np.abs(f.df)) < 1.0e-20 and (
            f.dfvar.size > 1 and np.max(np.abs(np.diff(f.dfvar))) < 1.0e-20
        ):
            print(
                "\nWarning: fitderiv may have failed for"
                f" {experiment}: {strain} in {condition}."
            )
        df_for_s, dict_for_sc, peak_coords = (
            omstats.find_smoothed_data_summary_stats(
                fitvar=fitvar,
                derivname=derivname,
                statnames=statnames,
                nosamples=nosamples,
                f=f,
                t=t,
                e=experiment,
                c=condition,
                s=strain,
                findareas=findareas,
                plotlocalmax=plotlocalmax,
                printpeakproperties=printpeakproperties,
                **kwargs,
            )
        )
        if peak_coords is not None:
            f.peak_coords = peak_coords
        # store GP parameters
        dict_for_sc[f"logmaxlikehood_for_{derivname}"] = f.logmaxlike
        dict_for_sc["gp_for_" + derivname] = cvfn
        for j, val in enumerate(f.lth):
            dict_for_sc[f"log_hyperparameter_{j}_for_{derivname}"] = val
        f.df_for_s = df_for_s
        f.dict_for_sc = dict_for_sc
        return f
    else:
        raise Exception("fitderiv failed.")


class fitderiv:
    """
    Smooth and estimate the time derivative of the data via Gaussian processes.

    After a successful optimisation, the following attributes are generated:

    t: array
        The times specified as input.
    d: array
        The data specified as input.
    f: array
        The mean of the Gaussian process with the optimal hyperparmeters
        at each time point.
    fvar: array
        The variance of the optimal Gaussian process at each time point.
    df: array
        The inferred first time-derivative.
    dfvar: array
        The inferred variance of the first time-derivative.
    ddf: array
        The inferred second time-derivative.
    ddfvar: array
        The inferred variance of the second time-derivative.

    Examples
    --------
    A typical work flow is:

    >>> from fitderiv import fitderiv
    >>> q= fitderiv(t, od)
    >>> q.plotfit('df')

    or potentially

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(q.t, q.d, 'r.', q.t, q.y, 'b')

    Reference
    ---------
    PS Swain, K Stevenson, A Leary, LF Montano-Gutierrez, IBN Clark, J Vogel,
    and T Pilizota.
    Inferring time derivatives including growth rates using Gaussian processes
    Nat Commun 7 (2016) 13766
    """

    def __init__(
        self,
        t,
        d,
        cvfn="matern",
        logs=True,
        negs2nan=True,
        noruns=5,
        noinits=100,
        exitearly=False,
        bd=None,
        empirical_errors=False,
        optmethod="L-BFGS-B",
        runtime_warn=False,
        linalgmax=3,
        maxdatapts=None,
        warn=False,
    ):
        """
        Smooth data and estimate time derivatives with a Gaussian process.

        Parameters
        ----------
        t: 1D array
            The time points.
        d: array
            The data corresponding to the time points with any replicates given
             as columns.
        cvfn: string
            The type of kernel function for the Gaussian process either 'sqexp'
            (squared exponential) or 'matern' (Matern with nu= 5/2) or 'nn'
            (neural network).
        logs: boolean
            If True, the Gaussian process is used to smooth the natural
            logarithm of the data and the time-derivative is therefore of the
            logarithm of the data.
        negs2nan: boolean, optional
            If logs=True, negs2nan=True causes all negative numbers to be set
            to np.nan; negs2nan=False, causes all negative numbers to be set
            almost zero.
        noruns: integer, optional
            The number of attempts to be made at optimising the kernel's
            hyperparmeters.
        noinits: integer, optional
            The number of random attempts made to find good initial choices for
            the hyperparameters before running their optimisation.
        exitearly: boolean, optional
            If True, stop at the first successful attempt at optimising the
            hyperparameters otherwise take the best choice from all successful
            optimisations.
        bd: dictionary, optional
            Specifies the limits on the hyperparameters for the Gaussian process.
            For example, bd= {0: [-1, 4], 2: [2, 6]})
            sets confines the first hyperparameter to be between 1e-1 and 1e^4
            and confines the third hyperparmater between 1e2 and 1e6.
        empirical_errors: boolean, optional
            If True, measurement errors are empirically estimated by the
            variance across replicates at each time point.
            If False, the variance of the measurement error is assumed to be
            the same for all time points and its magnitude is a hyperparameter
            that is optimised.
        optmethod: string, optional
            The algorithm used to optimise the hyperparameters, either
            'l_bfgs_b' or 'tnc'.
        warn: boolean, optional
            If False, warnings created by covariance matrices that are not
            positive semi-definite are suppressed.
        linalgmax: integer, optional
            The number of times errors generated by underlying linear algebra
            modules during the optimisation by poor choices of the
            hyperparameters should be ignored.
        maxdatapts: integer, optional
            If set, sufficiently large data sets with multiple replicates will
            be subsampled at each time point, randomly picking a smaller
            number of replicates, to reduce the number of data points and so
            run times.
        """
        self.linalgmax = linalgmax
        if not runtime_warn:
            # warning generated occasionally when sampling from the Gaussian
            # process likely because of numerical errors
            warnings.simplefilter("ignore", RuntimeWarning)
        try:
            noreps = d.shape[1]
        except IndexError:
            noreps = 1
        self.d = np.copy(d)
        self.t = np.copy(t)
        t = self.t
        d = self.d
        # default bounds for hyperparameters
        bddict = {
            "nn": {0: (-1, 5), 1: (-7, -2), 2: (-6, 2)},
            "sqexp": {0: (-5, 5), 1: (-6, 2), 2: (-5, 2)},
            "matern": {0: (-5, 5), 1: (-4, 4), 2: (-2, 0)},
        }
        # find bounds
        if bd:
            bds = bddict[cvfn] | bd
        else:
            bds = bddict[cvfn]
        # log data
        if logs:
            print("Taking natural logarithm of the data.")
            if np.nonzero(d <= 0)[0].size > 0:
                if negs2nan:
                    print("Warning: Negative data is being set to NaN.")
                    # replace zeros and negs so that logs can be applied
                    d[np.nonzero(d <= 0)] = np.nan
                else:
                    # NB too small values can generate odd samples from the GPs
                    print("Warning: Negative data is being set close to zero.")
                    d[np.nonzero(d <= 0)] = np.finfo(float).eps
            # take log of data
            d = np.log(np.asarray(d))
        # run checks and define measurement errors
        if empirical_errors:
            # errors must be empirically estimated
            if noreps > 1:
                lod = [
                    np.count_nonzero(np.isnan(d[:, i])) for i in range(noreps)
                ]
                if np.sum(np.diff(lod)) != 0:
                    print(
                        "The replicates have different number of data "
                        "points, but equal numbers of data points are "
                        "needed for empirically estimating errors."
                    )
                    merrors = None
                else:
                    # estimate errors empirically
                    print("Estimating measurement errors empirically.")
                    merrors = gu.findsmoothvariance(d)
            else:
                print("Not enough replicates to estimate errors empirically.")
                merrors = None
        else:
            merrors = None
        ta, da, ma, preprocessing_success = preprocess_data(
            t, d, merrors, maxdatapts
        )
        if preprocessing_success:
            self.run(
                cvfn=cvfn,
                bds=bds,
                ta=ta,
                da=da,
                ma=ma,
                noruns=noruns,
                noinits=noinits,
                exitearly=exitearly,
                optmethod=optmethod,
            )

    def run(
        self,
        cvfn,
        bds,
        ta,
        da,
        ma,
        noruns,
        noinits,
        exitearly,
        optmethod,
    ):
        """Instantiate and run a Gaussian process."""
        try:
            # instantiate GP
            g = getattr(gp, cvfn + "GP")(bds, ta, da, merrors=ma)
            print("Using a " + g.description + ".")
        except NameError:
            raise Exception("Gaussian process not recognised.")
        try:
            # optimise parameters
            g.findhyperparameters(
                noruns,
                noinits=noinits,
                exitearly=exitearly,
                optmethod=optmethod,
                linalgmax=self.linalgmax,
            )
            # display results
            if np.any(ma):
                # check measurement errors
                if len(ma) != len(self.t):
                    # NaNs have been removed
                    mainterp = interp1d(
                        ta, ma, bounds_error=False, fill_value=(ma[0], ma[-1])
                    )
                    ma = mainterp(self.t)
            g.predict(self.t, derivs=2, addnoise=True, merrorsnew=ma)
            self.success = True
            # results
            self.gp = g
            self.logmaxlike = -g.nlml_opt
            self.hparamerr = g.hparamerr
            self.lth = g.lth_opt
            self.f = g.f
            self.df = g.df
            self.ddf = g.ddf
            self.fvar = g.fvar
            self.dfvar = g.dfvar
            self.ddfvar = g.ddfvar
        except gaussianprocessException:
            self.success = False

    def fitderivsample(self, nosamples, newt=None):
        """
        Generate samples from the latent function.

        Both values for the latent function and its first two
        derivatives are returned, as a tuple.

        All derivatives must be sampled because by default all are asked
        to be predicted by the underlying Gaussian process.

        Parameters
        ----------
        nosamples: integer
            The number of samples.
        newt: array, optional
            Time points for which the samples should be made.
            If None, the orginal time points are used.

        Returns
        -------
        samples: a tuple of arrays
            The first element of the tuple gives samples of the latent
            function;
            the second element gives samples of the first time derivative; and
            the third element gives samples of the second time derivative.
        """
        if np.any(newt):
            newt = np.asarray(newt)
            # make prediction for new time points
            gps = copy.deepcopy(self.gp)
            gps.predict(newt, derivs=2, addnoise=True)
        else:
            gps = self.gp
        samples = gps.sample(nosamples, derivs=2)
        return samples

    def plotfit(
        self,
        experiment,
        condition,
        strain,
        fitvar,
        derivname,
        logs,
    ):
        """Create figure using subplotfit to generate subplots."""
        figtitle = f"{experiment}: {strain} in {condition}"
        plt.figure()
        plt.subplot(2, 1, 1)
        self.subplotfit("f", ylabel=fitvar, figtitle=figtitle, logs=logs)
        plt.subplot(2, 1, 2)
        self.subplotfit("df", logs=logs, ylabel=derivname)
        # add dot on largest local peak
        if hasattr(self, "peak_coords"):
            plt.plot(
                self.peak_coords[0],
                self.peak_coords[1],
                "o",
                color="yellow",
                markeredgecolor="k",
            )
        plt.tight_layout()
        plt.show(block=False)

    def subplotfit(
        self,
        char="f",
        errorfac=1,
        logs=None,
        xlabel="time",
        ylabel=False,
        figtitle=False,
    ):
        """
        Plot the results of the fitting.

        Either the data and the mean of the optimal Gaussian process or
        the inferred time derivatives are plotted.

        Parameters
        ----------
        char: string
            The variable to plot either 'f' or 'df' or 'ddf'.
        errorfac: float, optional
            The size of the errorbars are errorfac times the standard deviation
            of the optimal Gaussian process.
        ylabel: string, optional
            A label for the y-axis.
        figtitle: string, optional
            A title for the figure.
        """
        x = getattr(self, char)
        xv = getattr(self, char + "var")
        if char == "f":
            if logs:
                d = np.log(self.d)
            else:
                d = self.d
            plt.plot(self.t, d, "r.")
        plt.plot(self.t, x, "b")
        plt.fill_between(
            self.t,
            x - errorfac * np.sqrt(xv),
            x + errorfac * np.sqrt(xv),
            facecolor="blue",
            alpha=0.2,
        )
        if ylabel:
            plt.ylabel(ylabel)
        else:
            plt.ylabel(char)
        plt.xlabel(xlabel)
        if figtitle:
            plt.title(figtitle)


def subsample_data(t, d, no_samples=2000):
    """
    Subsample replicate data.

    At each time point, randomly choose only some of the replicates.
    """
    rng = np.random.default_rng()
    ta = np.repeat(t.reshape(-1, 1), d.shape[1], axis=1)
    rd = d[np.isfinite(d)]
    ta = ta[np.isfinite(d)]
    # make dict with time points as keys and corresponding d as values
    t_dict = {tv: rd[np.where(ta == tv)[0]] for tv in t}
    no_samples_per_tpt = np.max([1, int(no_samples / t.size)])
    # subsample
    nd = np.array(
        [
            (
                rng.choice(t_dict[tv], no_samples_per_tpt, replace=False)
                if no_samples_per_tpt < t_dict[tv].size
                else rng.choice(t_dict[tv], no_samples_per_tpt, replace=True)
            )
            for tv in t_dict
        ]
    )
    return nd


def preprocess_data(t, d, merrors, maxdatapts):
    """Remove nans and make 1D."""
    try:
        noreps = d.shape[1]
    except IndexError:
        noreps = 1
    # subsample if excessive data
    rd = d[~np.isnan(d)]
    if maxdatapts and rd.size > maxdatapts:
        print(
            f"Many data points - {rd.size}: subsampling for each time point."
        )
        nd = subsample_data(t, d, maxdatapts)
        print(f"Using {nd.size} data points.")
    else:
        nd = d
    # combine data into one array
    tb = np.tile(t, noreps)
    db = np.reshape(nd, nd.size, order="F")
    # check for NaNs
    if np.any(merrors):
        mb = np.tile(merrors, noreps)
        keep = np.intersect1d(
            np.nonzero(~np.isnan(db))[0], np.nonzero(~np.isnan(mb))[0]
        )
    else:
        keep = np.nonzero(~np.isnan(db))[0]
    # remove any NaNs
    da = db[keep]
    ta = tb[keep]
    # check data remains after removing NaNs
    success = True
    if not da.size:
        print("Warning: omfitderiv failed - too many NaNs.")
        success = False
    elif np.any(merrors):
        # measurement errors
        ma = mb[keep]
        if not ma.size:
            print("Warning: omfitderiv failed - too many NaNs.")
            success = False
    else:
        ma = None
    return ta, da, ma, success



import numpy as np

import omniplate.corrections as omcorr
import omniplate.sunder as sunder
from omniplate.omfitderiv import runfitderiv


def getpromoteractivity(
    self,
    f="GFP",
    figs=True,
    flcvfn="matern",
    bd=None,
    nosamples=1000,
    maxdatapts=None,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
    strains="all",
    strainincludes=False,
    strainexcludes=False,
):
    """
    Estimate the promoter activity from the corrected fluorescence.

    A Gaussian process is run on the corrected fluorescence, and
    the results from getstats are used.

    Arguments
    --
    f: string
        The fluorescence measurement, typically either 'GFP or 'mCherry'.
    figs: boolean
        If True, display plots showing the fits to the corrected fluorescence.
    flcvfn: str, optional
        The covariance function to use for the Gaussian process applied
        to the logarithm of the corrected fluorescence.
    bd: dict, optional
        Specifies the bounds on the hyperparameters for the Gaussian
        process applied to the logarithm of the corrected fluorescence,
        e.g. {2: (-2, 0)}.
    nosamples: int, optional
        The number of samples to take when using Gaussian processes.
    maxdatapts: int, optional
        The maximum number of data points to use for the Gaussian process.
        Too many data points, over 1500, can be slow.
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in
        their name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.
    """
    for e in sunder.getset(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        "experiment",
        nonull=True,
    ):
        for c in sunder.getset(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            labeltype="condition",
            nonull=True,
            nomedia=True,
        ):
            for s in sunder.getset(
                self,
                strains,
                strainincludes,
                strainexcludes,
                labeltype="strain",
                nonull=True,
            ):
                if f"{s} in {c}" in self.allstrainsconditions[e]:
                    choose = (
                        (self.s.experiment == e)
                        & (self.s.condition == c)
                        & (self.s.strain == s)
                    )
                    if f"c{f}" in self.s.columns:
                        cfl = self.s[choose][f"c{f}"].values
                        cfls = np.array(
                            [
                                cfl,
                                cfl + self.s[choose][f"c{f}_err"].values,
                                cfl - self.s[choose][f"c{f}_err"].values,
                            ]
                        ).T
                        t = self.s[choose].time.values
                        if cfl.size == 0:
                            print(
                                f"Warning: No corrected {f} found for"
                                f" {e}: {s} in {c}."
                            )
                            continue
                        fitvar = f"log_c{f}"
                        derivname = f"d/dt_{fitvar}"
                        f_fitderiv = runfitderiv(
                            t=t,
                            d=cfls,
                            fitvar=fitvar,
                            derivname=derivname,
                            experiment=e,
                            condition=c,
                            strain=s,
                            bd=bd,
                            cvfn=flcvfn,
                            logs=True,
                            maxdatapts=maxdatapts,
                        )
                        if not f_fitderiv.success:
                            print(
                                "\n-> Fitting corrected fluorescence failed"
                                f" for {e}: {s} in {c}.\n"
                            )
                            continue
                        if figs:
                            f_fitderiv.plotfit(
                                e, c, s, fitvar, derivname, logs=True
                            )
                        # samples
                        t_od, od = sunder.extractwells(
                            self.r, self.s, e, c, s, ["OD"]
                        )
                        lod_samples = omcorr.sample_ODs_with_GP(
                            self, e, c, s, t_od, od, nosamples
                        )
                        lfl_samples, lfl_deriv_samples = (
                            f_fitderiv.fitderivsample(nosamples)[:2]
                        )
                        # promoter activity samples
                        pa_samples = lfl_deriv_samples * np.exp(
                            lfl_samples - lod_samples
                        )
                        # store results
                        resdict = {
                            "experiment": e,
                            "condition": c,
                            "strain": s,
                            "time": t,
                            f"promoter_activity_{f}_err": omcorr.naniqrzeros2nan(
                                pa_samples, 1
                            ),
                            f"promoter_activity_{f}": np.nanmean(
                                pa_samples, 1
                            ),
                        }
                        # add to data frames
                        omcorr.addtodataframes(
                            self, resdict, "promoter_activity_{f}"
                        )
                else:
                    print(
                        "No corrected fluorescence found for "
                        f"{e}: {s} in {c}."
                    )



"""Display information on the data sets and on omniplate."""

import pprint
from pathlib import Path

import omniplate.omgenutils as gu


@property
def info(self):
    """
    Display conditions, strains, and data types.

    Example
    -------
    >>> p.info
    """
    if self.allexperiments:
        for exp in self.allexperiments:
            print("\nExperiment:", exp, "\n---")
            print("Conditions:")
            for c in sorted(self.allconditions[exp], key=gu.natural_keys):
                print("\t", c)
            print("Strains:")
            for s in sorted(self.allstrains[exp], key=gu.natural_keys):
                print("\t", s)
            print("Data types:")
            for d in self.datatypes[exp]:
                print("\t", d)
            if self.progress["ignoredwells"]:
                print("Ignored wells:")
                if self.progress["ignoredwells"][exp]:
                    for d in self.progress["ignoredwells"][exp]:
                        print("\t", d)
                else:
                    print("\t", "None")
        print()
    else:
        print("No experiments loaded.")


@property
def find_available_data(self):
    """Create files and data sets as attributes."""
    files = []
    datasets = []
    for f in self.datadirpath.glob("*.*"):
        if f.is_file() and (
            f.suffix == ".xlsx"
            or f.suffix == ".json"
            or f.suffix == ".tsv"
            or f.suffix == ".csv"
            or f.suffix == ".xls"
        ):
            files.append(f.stem + f.suffix)
            if f.suffix == ".tsv" or f.suffix == ".json" or f.suffix == ".csv":
                froot = "_".join(f.stem.split("_")[:-1])
                if froot not in datasets:
                    datasets.append(froot)
    self.files = {i: f for i, f in enumerate(sorted(files))}
    self.datasets = {i: f for i, f in enumerate(sorted(datasets))}


@property
def ls(self):
    """
    List all files in the data directory.

    A dictionary of available files to load and a list of available
    data sets to import are created as a shortcuts.

    Parameter
    --------
    output: boolean
        If True, list available files.

    Examples
    --------
    >>> p.ls
    >>> p.files
    >>> p.load(p.files[0], p.files[1])
    >>> p.importdf(p.datasets)
    """
    print(f"Data directory is {str(self.datadirpath.resolve())}.")
    print(f"Working directory is {str(self.wdirpath.resolve())}.")
    print("Files available - see .files - are:", "\n---")
    pprint.pprint(self.files)
    print()


def changewdir(self, wdir, ls=True):
    """
    Change working directory.

    Parameters
    ----------
    wdir: string
        The new working directory specified from the current directory.
    ls: boolean
        If True (default), display contents of the working directory.

    Example
    -------
    >>> p.changewdir('newdata/')
    """
    self.wdirpath = Path(wdir)
    self.find_available_data
    if ls:
        self.ls


def changedatadir(self, datadir, ls=True):
    """
    Change data directory.

    Parameters
    ----------
    datadir: string
        The new working directory specified from the current directory.
    ls: boolean
        If True (default), display contents of the working directory.

    Example
    -------
    >>> p.changewdir('newdata/')
    """
    self.datadirpath = Path(datadir)
    self.find_available_data
    if ls:
        self.ls


@property
def webhelp(self):
    """
    Open detailed examples of how to use in omniplate in a web browser.

    Example
    -------
    >>> p.webhelp
    """
    import webbrowser

    url = "https://swainlab.bio.ed.ac.uk/software/omniplate/index.html"
    webbrowser.get(None).open_new(url)


@property
def log(self):
    """
    Print a log of all methods called and their arguments.

    Example
    -------
    >>> p.log
    """
    print(self.logstream.getvalue())


@property
def experiment_map(self):
    """For a combined experiment, show the original experiments and IDs."""
    if hasattr(self, "_exp_map"):
        return self._exp_map
    else:
        if "experiment_id" and "original_experiment" in self.r.columns:
            self._exp_map = {
                name: int(id)
                for id, name in zip(
                    self.r.experiment_id.unique(),
                    self.r.original_experiment.unique(),
                )
            }
            return self._exp_map
        else:
            print("There is only one experiment.")



"""Find and analyse mid-log growth."""

import matplotlib.pylab as plt
import numpy as np
from nunchaku import Nunchaku

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omgenutils as gu
import omniplate.sunder as sunder


@clogger.log
def getmidlog(
    self,
    stats=["mean", "median", "min", "max"],
    min_duration=1,
    max_num=4,
    prior=[-5, 5],
    use_smoothed=False,
    figs=True,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
    strains="all",
    strainincludes=False,
    strainexcludes=False,
    **kwargs,
):
    """
    Calculate mid-log statistics.

    Find the region of mid-log growth using nunchaku and calculate a
    statistic for each variable in the s dataframe in this region only.

    The results are added to the sc dataframe.

    Parameters
    ----------
    stats: str, optional
        A list of statistics to be calculated (using pandas).
    min_duration: float, optional
        The expected minimal duration of the midlog phase in units of time.
    max_num: int, optional
        The maximum number of segments of a growth curve.
    prior: list of two floats, optional
        Prior for nunchaku giving the lower and upper bounds on the gradients of the line segments.
    use_smoothed: boolean, optional
        If True, use the smoothed OD found by getstats and its estimated
        errors.
        If False, use the OD of the replicates in different wells.
    figs: boolean, optional
        If True, show nunchaku's results with the mid-log region marked by
        black squares.
    experiments: string or list of strings, optional
        The experiments to include.
    conditions: string or list of strings, optional
        The conditions to include.
    strains: string or list of strings, optional
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in
        their name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.
    kwargs: passed to Nunchaku
    """
    admin.check_kwargs(kwargs)
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull=True,
        nomedia=True,
    )
    # run Nunchaku to find midlog and take the mean of summary stats
    find_midlog_stats(
        self,
        stats,
        min_duration,
        max_num,
        prior,
        use_smoothed,
        figs,
        exps,
        cons,
        strs,
        **kwargs,
    )


def find_midlog_stats(
    self,
    stats,
    min_duration,
    max_num,
    prior,
    use_smoothed,
    figs,
    exps,
    cons,
    strs,
    **kwargs,
):
    """Find the stat of all variables in the s dataframe for mid-log growth."""
    stats = gu.makelist(stats)
    for e in exps:
        for c in cons:
            for s in strs:
                select = (
                    (self.s.experiment == e)
                    & (self.s.condition == c)
                    & (self.s.strain == s)
                )
                # get OD data
                if use_smoothed:
                    try:
                        Y = self.s[select]["smoothed_log_OD"].values
                        t = self.s[select]["time"].values
                        sd = self.s[select]["smoothed_log_OD_err"].values
                    except KeyError:
                        print(
                            f"Warning: smoothed ODs do not exist for {e}:{s}"
                            " in {c}"
                        )
                        continue
                else:
                    t, od = sunder.extractwells(self.r, self.s, e, c, s, "OD")
                    if np.any(t):
                        od[od < 0] = np.finfo(float).eps
                        Y = np.log(od.T)
                    else:
                        Y = []
                if len(Y):
                    print(f"\nFinding mid-log growth for {e} : {s} in {c}")
                    # run nunchaku on log(OD)
                    if use_smoothed:
                        # use standard deviation from GP
                        err = sd
                    else:
                        # estimate standard deviation from replicates
                        err = None
                    # find time points with no NaN
                    common_i = np.all(~np.isnan(Y), axis=0)
                    try:
                        # run nunchaku
                        nc = Nunchaku(
                            t[common_i],
                            Y[:, common_i],
                            err=err,
                            prior=prior,
                            **kwargs,
                        )
                        num_segs, evidence = nc.get_number(num_range=max_num)
                        bds, bds_std = nc.get_iboundaries(num_segs)
                        res_df = nc.get_info(bds)
                    except OverflowError:
                        print(
                            f"\nError finding midlog data for {e}: {s} in {c}."
                        )
                        continue
                    # pick midlog segment
                    t_st, t_en = pick_midlog(res_df, min_duration)
                    if np.isnan(t_st) or np.isnan(t_en):
                        print(
                            f"\nError finding midlog data for {e}: {s} in {c}."
                        )
                        continue
                    # plot nunchaku's results
                    if figs:
                        nc.plot(res_df, hlmax=None)
                        for tv in [t_st, t_en]:
                            iv = np.argmin((t - tv) ** 2)
                            if use_smoothed:
                                y_med = Y
                            else:
                                y_med = np.median(Y, axis=0)
                            plt.plot(
                                tv,
                                y_med[iv],
                                "ks",
                                markersize=10,
                            )
                        plt.xlabel("time")
                        plt.ylabel("log(OD)")
                        plt.title(f"mid-log growth for {e} : {s} in {c}")
                        plt.show(block=False)
                    # find statistics over mid-log growth
                    add_midlog_stats_to_sc(
                        self, select, e, c, s, t_st, t_en, stats
                    )


def add_midlog_stats_to_sc(self, select, e, c, s, t_st, t_en, stats):
    """Find and store stats from midlog growth."""
    sdf = self.s[select]
    midlog_sdf = sdf[(sdf.time >= t_st) & (sdf.time <= t_en)]
    for stat in stats:
        # store results in dict
        res_dict = {col: np.nan for col in self.s.columns}
        res_dict["experiment"] = e
        res_dict["condition"] = c
        res_dict["strain"] = s
        stat_res = getattr(midlog_sdf, stat)(numeric_only=True)
        for key, value in zip(stat_res.index, stat_res.values):
            res_dict[key] = value
        # add "midlog" to data names
        res_dict = {
            (
                f"{stat}_midlog_{k}"
                if k
                not in [
                    "experiment",
                    "condition",
                    "strain",
                ]
                else k
            ): v
            for k, v in res_dict.items()
        }
        # add to sc dataframe
        admin.add_dict_to_sc(self, res_dict)


def pick_midlog(res_df, min_duration):
    """Find midlog from nunchaku's results dataframe."""
    # midlog had a minimal duration and positive growth rate
    sdf = res_df[(res_df["delta x"] > min_duration) & (res_df.gradient > 0)]
    if sdf.empty:
        return np.nan, np.nan
    else:
        # find mid-log growth - maximal specific growth rate
        ibest = sdf.index[sdf.gradient.argmax()]
        t_st, t_en = sdf["x range"][ibest]
        return t_st, t_en



# errors
class omniplateError(ValueError):
    pass


class FileNotFound(omniplateError):
    pass


class IgnoreWells(omniplateError):
    pass


class PlotError(omniplateError):
    pass


class UnknownDataFrame(omniplateError):
    pass


class getsubset(omniplateError):
    pass


class GetFitnessPenalty(omniplateError):
    pass


class CorrectAuto(omniplateError):
    pass


class UnknownPlateReader(omniplateError):
    pass



"""Function to parse data file generated by the plate reader."""

from pathlib import Path

import pandas as pd

import omniplate.omerrors as errors
from omniplate.parseSunrise import parseSunrise
from omniplate.parseTecan import parseTecan


def parseplate(
    platereaderfile,
    platereadertype,
    datadirpath=".",
    sheetnumber=0,
    export=False,
):
    """
    Parse plate-reader output into a long dataframe.

    Note that the plate-reader file is assumed to be an Excel file
    and that time is converted into hours.

    Parameters
    ----------
    platereaderfile: str
        The name of the data file created by a platereader.
    platereadertype: str
        The type of plate reader, currently only "Tecan"
        for a Tecan M200 or Tecan F200.
    datadirpath: str, optional
        The path to the platereader_file.
    sheetnumber: integer, optional
        The sheet to read from an excel file.
    export: boolean, optional
        If True, write parsed data to tsv file.

    Example
    -------
    >>> from om_code.parseplate import parseplate
    >>> rdf= parseplate("ExampleData.xlsx", "Tecan", datadirpath="data")
    >>> print(rdf)

                time well      OD   GFP  AutoFL  mCherry
    0       0.000000   A1  0.2555  46.0    18.0     19.0
    1       0.232306   A1  0.2725  45.0    17.0     17.0

    """
    if isinstance(datadirpath, str):
        datadirpath = Path(datadirpath)
    if platereadertype == "tidy":
        print(
            "Columns must be labelled 'time', 'well', 'OD', etc., "
            "and time must be in units of hours."
        )
        try:
            if ".tsv" in platereaderfile:
                rdf = pd.read_csv(
                    str(datadirpath / platereaderfile), sep="\t", index_col=0
                )
            else:
                rdf = pd.read_csv(
                    str(datadirpath / platereaderfile), index_col=0
                )
        except FileNotFoundError:
            raise errors.FileNotFound(str(datadirpath / platereaderfile))
        if rdf.time.max() > 100:
            print("Warning: time does not appear to be in hours.")
        return rdf
    else:
        # load data
        try:
            dfd = pd.read_excel(
                str(datadirpath / platereaderfile), sheet_name=sheetnumber
            )
        except FileNotFoundError:
            raise errors.FileNotFound(str(datadirpath / platereaderfile))
        # create a dict to store data
        rdict = {"time": [], "well": []}
        # parse loaded data frame
        if platereadertype == "Tecan":
            rdict = parseTecan(dfd, rdict)
        elif platereadertype == "Sunrise":
            rdict = parseSunrise(dfd, rdict)
        else:
            raise ValueError(f"{platereadertype} not recognised.")
        # convert parsed dict to data frame
        rdf = pd.DataFrame.from_dict(rdict)
        if export:
            outname = platereaderfile.split(".")[0] + ".tsv"
            print(f"exporting {outname}")
            rdf.to_csv(str(datadirpath / outname), sep="\t")
        return rdf



"""Function to estimate growth rate."""

import concurrent.futures
import os

import numpy as np

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omplot as omplot
import omniplate.sunder as sunder
from omniplate.omfitderiv import runfitderiv


@clogger.log
def getstats(
    self,
    dtype="OD",
    noprocessors=1,
    figs=True,
    options=None,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
    strains="all",
    strainincludes=False,
    strainexcludes=False,
):
    """
    Smooth data, find its derivatives, and calculate summary statistics.

    The first and second time derivatives are found, typically of OD,
    using a Gaussian process (Swain et al., 2016).

    The derivatives are stored in the .s dataframe;
    summary statistics are stored in the .sc dataframe.

    Parameters
    ----------
    dtype: string, optional
        The type of data - 'OD', 'GFP', 'cGFPperOD', or 'cGFP' - for
        which the derivatives are to be found. The data must exist in the
        .r or .s dataframes.
    noprocessors: int
        Default is 1.
    figs: boolean, optional
        If True, plot both the fits and inferred derivative.
    options: None or dict
        The possible keys are:
            bd: dict
                The bounds on the hyperparameters for the Gaussian process.
                For example, bd= {1: [-2,0])} fixes the bounds on the
                hyperparameter controlling flexibility to be 1e-2 and 1e0.
                The default for a Matern covariance function
                is {0: (-5,5), 1: (-4,4), 2: (-5,2)},
                where the first element controls amplitude, the second controls
                flexibility, and the third determines the magnitude of the
                measurement error.
            cvfn: string
                The covariance function used in the Gaussian process, either
                'matern' or 'sqexp' or 'nn'.
            empirical_errors: boolean
                If True, measurement errors are empirically estimated from the
                variance across replicates at each time point and so vary with
                time.
                If False, the magnitude of the measurement error is fit from
                the data assuming that this magnitude is the same at all time
                points.
            noruns: integer
                The number of attempts made for each fit. Each attempt is made
                with random initial estimates of the hyperparameters within
                their bounds.
            exitearly: boolean
                If True, stop at the first successful fit.
                If False, use the best fit from all successful fits.
            noinits: integer
                The number of random attempts to find a good initial condition
                before running the optimization.
            nosamples: integer
                The number of samples used to calculate errors in statistics by
                bootstrapping.
            logs: boolean
                If True, find the derivative of the log of the data and should be
                True to determine the specific growth rate when dtype= 'OD'.
            findareas: boolean
                If True, find the area under the plot of gr vs OD and the area
                under the plot of OD vs time. Setting to True can make getstats
                slow.
            plotlocalmax: boolean
                If True, mark the highest local maxima found, which is used to
                calculate statistics, on any plots.
            printpeakproperties: boolean
                If True, show properties of any local peaks that have found by
                scipy's find_peaks. Additional properties can be specified as
                kwargs and are passed to find_peaks.
            maxdatapts: integer
                If set, sufficiently large data sets with multiple replicates
                will be subsampled at each time point, randomly picking a
                smaller number of replicates, to reduce the number of data
                points and so run times.
            Parameters for scipy's find_peaks can be include too.
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in their
        name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.

    Examples
    --------
    >>> p.getstats()
    >>> p.getstats(conditionincludes= "Gal", noprocessors=8)
    >>> p.getstats(options = {"noruns": 10, "exitearly": False})

    If the fits are poor, often changing the bounds on the hyperparameter
    for the measurement error helps:

    >>> p.getstats(options = {"bd": {2: (-3,0)}})

    References
    ----------
    PS Swain, K Stevenson, A Leary, LF Montano-Gutierrez, IB Clark,
    J Vogel, T Pilizota. (2016). Inferring time derivatives including cell
    growth rates using Gaussian processes. Nat Commun, 7, 1-8.
    """
    default_options = {
        "bd": None,
        "cvfn": "matern",
        "empirical_errors": False,
        "noruns": 10,
        "exitearly": True,
        "noinits": 100,
        "nosamples": 100,
        "logs": True,
        "findareas": False,
        "plotlocalmax": True,
        "printpeakproperties": False,
        "maxdatapts": None,
        "linalgmax": 5,
    }
    if options is None:
        options = default_options
    else:
        options = default_options | options
    # variable to be fit
    if options["logs"]:
        fitvar = f"log_{dtype}"
    else:
        fitvar = dtype
    # name of derivative of fit variable
    if fitvar == "log_OD":
        derivname = "gr"
    else:
        derivname = f"d/dt_{fitvar}"
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull=True,
    )
    getstats_many(
        self,
        noprocessors,
        figs,
        exps,
        cons,
        strs,
        dtype,
        fitvar,
        derivname,
        options,
    )


def getstats_many(
    self,
    noprocessors,
    figs,
    exps,
    cons,
    strs,
    dtype,
    fitvar,
    derivname,
    options,
):
    """Parallelize getstats."""
    params = {
        "fitvar": fitvar,
        "derivname": derivname,
        "fitderiv_params": options,
    }
    if noprocessors > 1:
        # multiple processors
        max_workers = np.min([os.cpu_count() - 1, noprocessors])
        print(f"Using {max_workers} processors out of {os.cpu_count()}.")
        results = run_many(self, exps, cons, strs, dtype, params, max_workers)
        # process results
        print("---")
        for e, c, s, f_fitderiv in results:
            print(f"Results for {e}: {s} in {c}")
            process_results_from_fitderiv(
                self, f_fitderiv, params, figs, e, c, s
            )

    else:
        # single processor
        for e in exps:
            for c in cons:
                for s in strs:
                    res = getdata_one(self, dtype, e, c, s)
                    if res is not None:
                        t, d = res
                        f_fitderiv = getstats_one(e, c, s, t, d, params)[-1]
                        process_results_from_fitderiv(
                            self, f_fitderiv, params, figs, e, c, s
                        )
        # print information on GP's hyperparameters
    f_fitderiv.gp.info


def run_many(self, exps, cons, strs, dtype, params, max_workers):
    """Submit getstats_one to multiple processors and return results."""
    results = []
    with concurrent.futures.ProcessPoolExecutor(
        max_workers=max_workers
    ) as executor:
        # create and submit all tasks
        futures = []
        for e in exps:
            for c in cons:
                for s in strs:
                    res = getdata_one(self, dtype, e, c, s)
                    if res is not None:
                        t, d = res
                        futures.append(
                            executor.submit(
                                getstats_one, e, c, s, t, d, params
                            )
                        )
        # collect results as they complete
        for future in concurrent.futures.as_completed(futures):
            try:
                result = future.result()
                if result is not None:
                    results.append(result)
            except Exception as exc:
                print(f"One processor generated an exception: {exc}")
    return results


def process_results_from_fitderiv(self, f_fitderiv, params, figs, e, c, s):
    """Add results from fitderiv to dataframes and plot."""
    # add time series to s dataframe
    admin.add_to_s(self, params["derivname"], f_fitderiv.df_for_s)
    # create or add summary stats to sc dataframe
    admin.add_dict_to_sc(self, f_fitderiv.dict_for_sc)
    if figs:
        f_fitderiv.plotfit(
            experiment=e,
            condition=c,
            strain=s,
            fitvar=params["fitvar"],
            derivname=params["derivname"],
            logs=params["fitderiv_params"]["logs"],
        )
    f_fitderiv.gp.results()
    print("---")


def getdata_one(self, dtype, e, c, s):
    """Get data for one condition and strain."""
    if f"{s} in {c}" not in self.allstrainsconditions[e]:
        return
    esc_name = f"{e}: {s} in {c}"
    if dtype in self.r.columns:
        # raw data
        t, d = sunder.extractwells(self.r, self.s, e, c, s, dtype)
    elif dtype in self.s.columns:
        # processed data
        df = self.s.query(
            "experiment == @e and condition == @c and strain == @s"
        )
        # add columns plus and minus err
        df = omplot.augmentdf(df, dtype)[[dtype, "augtype", "time"]]
        piv_df = df.pivot(index="time", columns="augtype", values=dtype)
        # convert to array for fitderiv
        d = piv_df.values
        t = piv_df.index.to_numpy()
        numberofnans = np.count_nonzero(np.isnan(d))
        if np.any(numberofnans):
            print(f"\nWarning: {numberofnans} NaNs in data")
    else:
        print(f"\n-> {dtype} not recognised for {esc_name}.\n")
        return
    # check data exists
    if d.size == 0:
        if esc_name.split(":")[1].strip() in self.allstrainsconditions[e]:
            print(f"\n-> No data found for {dtype} for {esc_name}.\n")
        return
    return t, d


def getstats_one(e, c, s, t, d, params):
    """Process a single getstats by running fitderiv."""
    f_fitderiv = runfitderiv(
        t=t,
        d=d,
        fitvar=params["fitvar"],
        derivname=params["derivname"],
        experiment=e,
        condition=c,
        strain=s,
        **params["fitderiv_params"],
    )
    return (e, c, s, f_fitderiv)



# !/usr/bin/env python

import matplotlib.pyplot as plt
import seaborn as sns

from omniplate.platereader import platereader

__version__ = "2.11.1"

plt.rcParams["figure.max_open_warning"] = 0
sns.set()
